@@ -2,7 +2,7 @@
  *      ipvsadm - IP Virtual Server ADMinistration program
  *                for IPVS NetFilter Module in kernel 2.4
  *
- *      Version: $Id: ipvsadm.c,v 1.22 2003/03/22 06:31:24 wensong Exp $
+ *      Version: $Id: ipvsadm.c,v 1.26 2004/01/10 09:39:03 wensong Exp $
  *
  *      Authors: Wensong Zhang <wensong@linuxvirtualserver.org>
  *               Peter Kese <peter.kese@ijs.si>
@@ -52,6 +52,8 @@
  *        Horms               :   added -v option
  *        Wensong Zhang       :   rewrite most code of parsing options and
  *                                processing options.
+ *        Alexandre Cassen    :   added ipvs_syncd SyncdID support to filter
+ *                                incoming sync messages.
  *
  *
  *      ippfvsadm - Port Fowarding & Virtual Server ADMinistration program
@@ -119,7 +121,7 @@
 #include "libipvs/libipvs.h"
 
 #define IPVSADM_VERSION_NO	"v" VERSION
-#define IPVSADM_VERSION_DATE	"2002/11/15"
+#define IPVSADM_VERSION_DATE	"2003/06/07"
 #define IPVSADM_VERSION		IPVSADM_VERSION_NO " " IPVSADM_VERSION_DATE
 
 #define MAX_TIMEOUT		(86400*31)	/* 31 days */
@@ -177,8 +179,10 @@ static const char* cmdnames[] = {
 #define OPT_STATS		0x04000
 #define OPT_RATE		0x08000
 #define OPT_THRESHOLDS		0x10000
-#define OPT_SORT		0x20000
-#define NUMBER_OF_OPT		18
+#define OPT_PERSISTENTCONN	0x20000
+#define OPT_SORT		0x40000
+#define OPT_SYNCID		0x80000
+#define NUMBER_OF_OPT		20
 
 static const char* optnames[] = {
 	"numeric",
@@ -198,7 +202,9 @@ static const char* optnames[] = {
 	"stats",
 	"rate",
 	"thresholds",
+	"persistent-conn",
 	"sort",
+	"syncid",
 };
 
 /*
@@ -211,43 +217,44 @@ static const char* optnames[] = {
  */
 static const char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] =
 {
-	/*   -n   -c   svc  -s   -p   -M   -r   fwd  -w   -x   -y   -mc  tot  dmn  -st  -rt  thr  srt */
-/*ADD*/     {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*EDIT*/    {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*DEL*/     {'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*FLUSH*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*LIST*/    {' ', '1', '1', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '1', '1', ' ', ' ', ' ', ' '},
-/*ADDSRV*/  {'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*DELSRV*/  {'x', 'x', '+', 'x', 'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*EDITSRV*/ {'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*TIMEOUT*/ {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*STARTD*/  {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', ' ', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*STOPD*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*RESTORE*/ {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*SAVE*/    {' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*ZERO*/    {'x', 'x', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+	/*   -n   -c   svc  -s   -p   -M   -r   fwd  -w   -x   -y   -mc  tot  dmn  -st  -rt  thr  -pc  srt  sid */
+/*ADD*/     {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', ' '},
+/*EDIT*/    {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*DEL*/     {'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*FLUSH*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*LIST*/    {' ', '1', '1', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '1', '1', ' ', ' ', ' ', ' ', ' ', 'x'},
+/*ADDSRV*/  {'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*DELSRV*/  {'x', 'x', '+', 'x', 'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*EDITSRV*/ {'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*TIMEOUT*/ {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*STARTD*/  {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', ' '},
+/*STOPD*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', ' '},
+/*RESTORE*/ {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*SAVE*/    {' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*ZERO*/    {'x', 'x', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
 };
 
 /* printing format flags */
-#define FMT_NONE	0x0000
-#define FMT_NUMERIC	0x0001
-#define FMT_RULE	0x0002
-#define FMT_STATS	0x0004
-#define FMT_RATE	0x0008
-#define FMT_THRESHOLDS	0x0010
-#define FMT_SORT	0x0020
+#define FMT_NONE		0x0000
+#define FMT_NUMERIC		0x0001
+#define FMT_RULE		0x0002
+#define FMT_STATS		0x0004
+#define FMT_RATE		0x0008
+#define FMT_THRESHOLDS		0x0010
+#define FMT_PERSISTENTCONN	0x0020
+#define FMT_SORT		0x0040
 
-#define SERVICE_NONE    0x0
-#define SERVICE_ADDR    0x1
-#define SERVICE_PORT    0x2
+#define SERVICE_NONE		0x0000
+#define SERVICE_ADDR		0x0001
+#define SERVICE_PORT		0x0002
 
 /* default scheduler */
-#define DEF_SCHED	"wlc"
+#define DEF_SCHED		"wlc"
 
 /* default multicast interface name */
-#define DEF_MCAST_IFN   "eth0"
+#define DEF_MCAST_IFN		"eth0"
 
-#define CONN_PROC_FILE  "/proc/net/ip_vs_conn"
+#define CONN_PROC_FILE		"/proc/net/ip_vs_conn"
 
 struct ipvs_command_entry {
 	int			cmd;
@@ -355,7 +362,7 @@ parse_options(int argc, char **argv, struct ipvs_command_entry *ce,
 		{"restore", 'R', POPT_ARG_NONE, NULL, 'R'},
 		{"save", 'S', POPT_ARG_NONE, NULL, 'S'},
 		{"start-daemon", '\0', POPT_ARG_STRING, &optarg, '1'},
-		{"stop-daemon", '\0', POPT_ARG_NONE, NULL, '2'},
+		{"stop-daemon", '\0', POPT_ARG_STRING, &optarg, '2'},
 		{"tcp-service", 't', POPT_ARG_STRING, &optarg, 't'},
 		{"udp-service", 'u', POPT_ARG_STRING, &optarg, 'u'},
 		{"fwmark-service", 'f', POPT_ARG_STRING, &optarg, 'f'},
@@ -373,11 +380,13 @@ parse_options(int argc, char **argv, struct ipvs_command_entry *ce,
 		{"numeric", 'n', POPT_ARG_NONE, NULL, 'n'},
 		{"connection", 'c', POPT_ARG_NONE, NULL, 'c'},
 		{"mcast-interface", '\0', POPT_ARG_STRING, &optarg, '3'},
+		{"syncid", '\0', POPT_ARG_STRING, &optarg, 'I'},
 		{"timeout", '\0', POPT_ARG_NONE, NULL, '5'},
 		{"daemon", '\0', POPT_ARG_NONE, NULL, '6'},
 		{"stats", '\0', POPT_ARG_NONE, NULL, '7'},
 		{"rate", '\0', POPT_ARG_NONE, NULL, '8'},
 		{"thresholds", '\0', POPT_ARG_NONE, NULL, '9'},
+		{"persistent-conn", '\0', POPT_ARG_NONE, NULL, 'P'},
 		{"sort", '\0', POPT_ARG_NONE, NULL, '0'},
 		{NULL, 0, 0, NULL, 0}
 	};
@@ -436,6 +445,11 @@ parse_options(int argc, char **argv, struct ipvs_command_entry *ce,
 		break;
 	case '2':
 		set_command(&ce->cmd, CMD_STOPDAEMON);
+		if (!strcmp(optarg, "master"))
+			ce->daemon.state = IP_VS_STATE_MASTER;
+		else if (!strcmp(optarg, "backup"))
+			ce->daemon.state = IP_VS_STATE_BACKUP;
+		else fail(2, "illegal start_daemon specified");
 		break;
 	case 'h':
 		usage_exit(argv[0], 0);
@@ -523,13 +537,13 @@ parse_options(int argc, char **argv, struct ipvs_command_entry *ce,
 		case 'x':
 			set_option(options, OPT_UTHRESHOLD);
 			if ((ce->dest.u_threshold =
-			     string_to_number(optarg, 0, 65535)) == -1)
+			     string_to_number(optarg, 0, INT_MAX)) == -1)
 				fail(2, "illegal u_threshold specified");
 			break;
 		case 'y':
 			set_option(options, OPT_LTHRESHOLD);
 			if ((ce->dest.l_threshold =
-			     string_to_number(optarg, 0, 65535)) == -1)
+			     string_to_number(optarg, 0, INT_MAX)) == -1)
 				fail(2, "illegal l_threshold specified");
 			break;
 		case 'c':
@@ -544,6 +558,12 @@ parse_options(int argc, char **argv, struct ipvs_command_entry *ce,
 			strncpy(ce->daemon.mcast_ifn,
 				optarg, IP_VS_IFNAME_MAXLEN);
 			break;
+		case 'I':
+			set_option(options, OPT_SYNCID);
+			if ((ce->daemon.syncid =
+			     string_to_number(optarg, 0, 255)) == -1)
+				fail(2, "illegal syncid specified");
+			break;
 		case '5':
 			set_option(options, OPT_TIMEOUT);
 			break;
@@ -562,6 +582,10 @@ parse_options(int argc, char **argv, struct ipvs_command_entry *ce,
 			set_option(options, OPT_THRESHOLDS);
 			*format |= FMT_THRESHOLDS;
 			break;
+		case 'P':
+			set_option(options, OPT_PERSISTENTCONN);
+			*format |= FMT_PERSISTENTCONN;
+			break;
 		case '0':
 			set_option(options, OPT_SORT);
 			*format |= FMT_SORT;
@@ -629,7 +653,7 @@ parse_options(int argc, char **argv, struct ipvs_command_entry *ce,
 		{"restore", 0, 0, 'R'},
 		{"set", 1, 0, '4'},
 		{"start-daemon", 1, 0, '1'},
-		{"stop-daemon", 0, 0, '2'},
+		{"stop-daemon", 1, 0, '2'},
 		{"tcp-service", 1, 0, 't'},
 		{"udp-service", 1, 0, 'u'},
 		{"fwmark-service", 1, 0, 'f'},
@@ -646,11 +670,13 @@ parse_options(int argc, char **argv, struct ipvs_command_entry *ce,
 		{"numeric", 0, 0, 'n'},
 		{"connection", 0, 0, 'c'},
 		{"mcast-interface", 1, 0, '3'},
+		{"syncid", 1, 0, 'I'},
 		{"timeout", 0, 0, '5'},
 		{"daemon", 0, 0, '6'},
 		{"stats", 0, 0, '7'},
 		{"rate", 0, 0, '8'},
 		{"thresholds", 0, 0, '9'},
+		{"persistent-conn", 0, 0, 'P'},
 		{"sort", 0, 0, '0'},
 		{"help", 0, 0, 'h'},
 		{0, 0, 0, 0}
@@ -727,6 +753,11 @@ parse_options(int argc, char **argv, struct ipvs_command_entry *ce,
 		break;
 	case '2':
 		set_command(&ce->cmd, CMD_STOPDAEMON);
+		if (!strcmp(optarg, "master"))
+			ce->daemon.state = IP_VS_STATE_MASTER;
+		else if (!strcmp(optarg, "backup"))
+			ce->daemon.state = IP_VS_STATE_BACKUP;
+		else fail(2, "illegal start_daemon specified");
 		break;
 	case 'h':
 		usage_exit(argv[0], 0);
@@ -812,19 +843,19 @@ parse_options(int argc, char **argv, struct ipvs_command_entry *ce,
 		case 'w':
 			set_option(options, OPT_WEIGHT);
 			if ((ce->dest.weight =
-			     string_to_number(optarg,0,65535)) == -1)
+			     string_to_number(optarg, 0, 65535)) == -1)
 				fail(2, "illegal weight specified");
 			break;
 		case 'x':
 			set_option(options, OPT_UTHRESHOLD);
 			if ((ce->dest.u_threshold =
-			     string_to_number(optarg, 0, 65535)) == -1)
+			     string_to_number(optarg, 0, INT_MAX)) == -1)
 				fail(2, "illegal u_threshold specified");
 			break;
 		case 'y':
 			set_option(options, OPT_LTHRESHOLD);
 			if ((ce->dest.l_threshold =
-			     string_to_number(optarg, 0, 65535)) == -1)
+			     string_to_number(optarg, 0, INT_MAX)) == -1)
 				fail(2, "illegal l_threshold specified");
 			break;
 		case 'c':
@@ -839,6 +870,12 @@ parse_options(int argc, char **argv, struct ipvs_command_entry *ce,
 			strncpy(ce->daemon.mcast_ifn,
 				optarg, IP_VS_IFNAME_MAXLEN);
 			break;
+		case 'I':
+			set_option(options, OPT_SYNCID);
+			if ((ce->daemon.syncid =
+			     string_to_number(optarg, 0, 255)) == -1)
+				fail(2, "illegal syncid specified");
+			break;
 		case '5':
 			set_option(options, OPT_TIMEOUT);
 			break;
@@ -857,6 +894,10 @@ parse_options(int argc, char **argv, struct ipvs_command_entry *ce,
 			set_option(options, OPT_THRESHOLDS);
 			*format |= FMT_THRESHOLDS;
 			break;
+		case 'P':
+			set_option(options, OPT_PERSISTENTCONN);
+			*format |= FMT_PERSISTENTCONN;
+			break;
 		case '0':
 			set_option(options, OPT_SORT);
 			*format |= FMT_SORT;
@@ -948,6 +989,7 @@ static int process_options(int argc, char **argv, int reading_stdin)
 		if ((options & OPT_CONNECTION ||
 		     options & OPT_TIMEOUT || options & OPT_DAEMON) &&
 		    (options & OPT_STATS ||
+		     options & OPT_PERSISTENTCONN ||
 		     options & OPT_RATE || options & OPT_THRESHOLDS))
 			fail(2, "options conflicts in the list command");
 
@@ -1012,7 +1054,7 @@ static int process_options(int argc, char **argv, int reading_stdin)
 		break;
 
 	case CMD_STOPDAEMON:
-		result = ipvs_stop_daemon();
+		result = ipvs_stop_daemon(&ce.daemon);
 	}
 
 	if (result)
@@ -1233,8 +1275,8 @@ static void usage_exit(const char *program, const int exit_status)
 		"  %s -L|l [options]\n"
 		"  %s -Z [-t|u|f service-address]\n"
 		"  %s --set tcp tcpfin udp\n"
-		"  %s --start-daemon state [--mcast-interface interface]\n"
-		"  %s --stop-daemon\n"
+		"  %s --start-daemon state [--mcast-interface interface] [--syncid sid]\n"
+		"  %s --stop-daemon state\n"
 		"  %s -h\n\n",
 		program, program, program,
 		program, program, program, program, program,
@@ -1277,12 +1319,14 @@ static void usage_exit(const char *program, const int exit_status)
 		"  --u-threshold  -x uthreshold        upper threshold of connections\n"
 		"  --l-threshold  -y lthreshold        lower threshold of connections\n"
 		"  --mcast-interface interface         multicast interface for connection sync\n"
+		"  --syncid sid                        syncid for connection sync (default=255)\n"
 		"  --connection   -c                   output of current IPVS connections\n"
 		"  --timeout                           output of timeout (tcp tcpfin udp)\n"
 		"  --daemon                            output of daemon information\n"
 		"  --stats                             output of statistics information\n"
 		"  --rate                              output of rate information\n"
 		"  --thresholds                        output of thresholds information\n"
+		"  --persistent-conn                   output of persistent connection info\n"
 		"  --sort                              sorting output of service/server entries\n"
 		"  --numeric      -n                   numeric output of addresses and ports\n",
 		DEF_SCHED);
@@ -1522,12 +1566,17 @@ static void print_title(unsigned int format)
 		       "  -> RemoteAddress:Port\n",
 		       "Prot LocalAddress:Port",
 		       "CPS", "InPPS", "OutPPS", "InBPS", "OutBPS");
-	else if (format & FMT_THRESHOLDS) {
+	else if (format & FMT_THRESHOLDS)
 		printf("%-33s %-10s %-10s %-10s %-10s\n"
 		       "  -> RemoteAddress:Port\n",
 		       "Prot LocalAddress:Port",
 		       "Uthreshold", "Lthreshold", "ActiveConn", "InActConn");
-	} else if (!(format & FMT_RULE))
+	else if (format & FMT_PERSISTENTCONN)
+		printf("%-33s %-9s %-11s %-10s %-10s\n"
+		       "  -> RemoteAddress:Port\n",
+		       "Prot LocalAddress:Port",
+		       "Weight", "PersistConn", "ActiveConn", "InActConn");
+	else if (!(format & FMT_RULE))
 		printf("Prot LocalAddress:Port Scheduler Flags\n"
 		       "  -> RemoteAddress:Port           Forward Weight ActiveConn InActConn\n");
 }
@@ -1620,7 +1669,7 @@ print_service_entry(ipvs_service_entry_t *se, unsigned int format)
 		if (!(dname = addrport_to_anyname(&daddr, ntohs(e->port),
 						  se->protocol, format)))
 			exit(1);
-		if ((!format & FMT_RULE))
+		if (!(format & FMT_RULE))
 			dname[28] = '\0';
 
 		if (format & FMT_RULE) {
@@ -1646,6 +1695,10 @@ print_service_entry(ipvs_service_entry_t *se, unsigned int format)
 			printf("  -> %-28s %-10u %-10u %-10u %-10u\n", dname,
 			       e->u_threshold, e->l_threshold,
 			       e->activeconns, e->inactconns);
+		} else if (format & FMT_PERSISTENTCONN) {
+			printf("  -> %-28s %-9u %-11u %-10u %-10u\n", dname,
+			       e->weight, e->persistconns,
+			       e->activeconns, e->inactconns);
 		} else
 			printf("  -> %-28s %-7s %-6d %-10u %-10u\n",
 			       dname, fwd_name(e->conn_flags),
@@ -1715,19 +1768,12 @@ static void list_daemon(void)
 	if (!(u = ipvs_get_daemon()))
 		exit(1);
 
-	switch (u->state) {
-	case IP_VS_STATE_NONE:
-		printf("not running\n");
-		break;
-	case IP_VS_STATE_MASTER:
-		printf("running as master (mcast=%s)\n", u->mcast_ifn);
-		break;
-	case IP_VS_STATE_BACKUP:
-		printf("running as backup (mcast=%s)\n", u->mcast_ifn);
-		break;
-	default:
-		fail(2, "invalid state");
-	}
+	if (u[0].state & IP_VS_STATE_MASTER)
+		printf("master sync daemon (mcast=%s, syncid=%d)\n",
+		       u[0].mcast_ifn, u[0].syncid);
+	if (u[1].state & IP_VS_STATE_BACKUP)
+		printf("backup sync daemon (mcast=%s, syncid=%d)\n",
+		       u[1].mcast_ifn, u[1].syncid);
 	free(u);
 }
 
